/*  libcalenabler - Library of calendaring and scheduling enabler
 *  eventbook.h - a set of functions associated with eventbook, which
 *                  is a collection of events
 *
 *  Authors: YANG Lin <lin.yang@orange-ftgroup.com> 
 *  
 *  This software and associated documentation files (the "Software") 
 *  are copyright (C) 2005 LiPS Linux Phone Standards Forum [FranceTelecom] 
 *  All Rights Reserved. 
 *
 *  A copyright license is hereby granted for redistribution and use of 
 *  the Software in source and binary forms, with or without modification, 
 *  provided that the following conditions are met: 
 *  - Redistributions of source code must retain the above copyright notice, 
 *  this copyright license and the following disclaimer. 
 *  - Redistributions in binary form must reproduce the above copyright 
 *   notice, this copyright license and the following disclaimer in the 
 *  documentation and/or other materials provided with the distribution. 
 *  - Neither the name of LiPS  nor the names of its Members may be used 
 *  to endorse or promote products derived from the Software without 
 *  specific prior written permission. 
 *
 *  A patent license for any Necessary Claims owned by Members of LiPS Forum 
 *  to make, have made, use, import, offer to sell, lease and sell or otherwise 
 *  distribute any implementation compliant with the any specification adopted 
 *  by the LiPS Forumcan be obtained from the respective Members on reasonable 
 *  and non-discriminatory terms and conditions and under reciprocity, as 
 *  regulated in more detail in the Internal Policy of the LiPS Forum. 
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER, ITS MEMBERS AND CONTRIBUTORS 
 *  "AS IS", AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE 
 *  AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER, 
 *  ITS MEMBERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *  POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef __LIPS_CAL_EVENTBOOK_H
#define __LIPS_CAL_EVENTBOOK_H

#include <calevent.h>

#ifdef __cplusplus
extern "C"
{
#endif                          /* __cplusplus */



/** Structure to hold cal_eventbook_t.
 * The instance structure _cal_eventbook_t must be defined in implementation source code.
 */
  typedef struct _cal_eventbook_t cal_eventbook_t;

/** @brief Connect the eventbook indicated by id
*
*  This function can be used to connect the eventbook indicated by id.
*
* @param    id             the eventbook id 
*
* @return   cal_eventbook_t if success;NULL if failure
*/
  cal_eventbook_t *eventbook_connect (cal_uint32_t id);

/** @brief Disconnect the eventbook 
*
*  This function can be used to disconnect the eventbook.
*
* @param    book             the eventbook 
*/
  void eventbook_disconnect (cal_eventbook_t *book);

/** @brief Add event to eventbook.
*
*  This function can be used to add a event to the data storage.
*
* @param    book           the eventbook to add event
* @param     event           the event to add to the eventbook
*
* @return   PIM_ERR_NONE   if success
* @return   cal_err_t      Error number if some error happen
*/
  cal_err_t eventbook_add_event (cal_eventbook_t *book, cal_event_t *event);

/** @brief Remove event from eventbook.
*
*  This function can be used to remove a event from the data storage.
*
* @param    book           the eventbook to remove event
* @param    event           the event to remove from the eventbook
*
* @return   PIM_ERR_NONE   if success
* @return   cal_err_t      Error number if some error happen
*/
  cal_err_t eventbook_remove_event (cal_eventbook_t *book, cal_char_t *uid);
/** @brief Remove more than one events from eventbook.
*
*  This function can be used to remove more than one events from the data storage.
*
* @param    book           the eventbook to remove event
* @param    event           the event to remove from the eventbook
*
* @return   PIM_ERR_NONE   if success
* @return   cal_err_t      Error number if some error happen
*
* @note      If nr_uids = 0, then all events will be removed.
*/
  cal_err_t eventbook_remove_events (cal_eventbook_t *book,
                                     cal_uint32_t nr_uids,
                                     const cal_char_t *uids[]);
/** @brief Remove all events from eventbook.
*
*  This function can be used to remove all events from the data storage.
*
* @param    book           the eventbook to remove event
*
* @return   PIM_ERR_NONE   if success
* @return   cal_err_t      Error number if some error happen
*/
  cal_err_t eventbook_remove_all_evnts (cal_eventbook_t *book);

/** @brief Update events to eventbook.
*
*  This function can be used to update event to the data storage.
*
* @param    book           the eventbook to update event
* @param    event           the event to update to the eventbook
*
* @return   PIM_ERR_NONE   if success
* @return   cal_err_t      Error number if some error happen
*/
  cal_err_t eventbook_update_event (cal_eventbook_t *book, cal_event_t *event);

/** @brief Get event according to uid
*
*  This function can be used to retrieve event according to event uid.
*
* @param    book           the eventbook to retrieve event from
* @param    event           the event retrieved from eventbook
*
* @return   PIM_ERR_NONE   if success
* @return   cal_err_t      Error number if some error happen
*/
  cal_err_t eventbook_get_event_by_uid (cal_eventbook_t *book, cal_event_t **event, 
                                        const cal_char_t *uid);

/** @brief Get all events
*
*  This function can be used to retrieve all events in an eventbook.
*
* @param    book           the eventbook to retrieve event from
*
* @return   PIM_ERR_NONE   if success
* @return   cal_err_t      Error number if some error happen
*/
  cal_err_t eventbook_get_all_events (cal_eventbook_t *book, cal_event_iter_t *iter);

/** @brief Get events during "from" to "to"
*
*  This function can be used to retrieve events during "from" to "to"
*
* @param    book           the eventbook to retrieve event 
* @param    iter          the cal_event_iter_t to store the events retrieved from eventbook
* @param    from           the start of the period
* @param    to             the end of the period
*
* @return   PIM_ERR_NONE   if success
* @return   cal_err_t      Error number if some error happen
* @note      If from =0 and to !=0, this function will get events before "to".
* @note      If from!=0 and to =0, this function will get events after "from".
* @note      If from =0 and to =0, this function will get all events.
* @note      If from !=0 and to !=0, this function will get events between "from" and "to".
*/
  cal_err_t eventbook_get_events_for_period (cal_eventbook_t *book,
                                             cal_event_iter_t *iter,
                                             cal_time_t from, cal_time_t to);



/** @brief skip a alarm
*
*  This function can be used to skip an alarm. If the event corresponding to the 
* uid is a repeating event, the next alarm time will be calculated and the new 
* alarm will be set.
*
* @param    book           the eventbook to retrieve event from
* @param    skip          the alarm time to skip
* @param    uid           the uid corresponding to the alarm 
*
* @return   PIM_ERR_NONE   if success
* @return   cal_err_t      Error number if some error happen
*/
  void schedule_next (cal_eventbook_t *book, cal_time_t skip, const cal_char_t *uid);
#ifdef __cplusplus
}
#endif                          /* __cplusplus */

#endif                          /* __LIPS_CAL_EVENTBOOK_H */

/*
vi:ts=2:nowrap:ai:expandtab
*/

