/*  libcalenabler - Library of calendaring and scheduling enabler
 *  
 *  Authors: YANG Lin <lin.yang@orange-ftgroup.com> 
 *  
 *  This software and associated documentation files (the "Software") 
 *  are copyright (C) 2005 LiPS Linux Phone Standards Forum [FranceTelecom] 
 *  All Rights Reserved. 
 *
 *  A copyright license is hereby granted for redistribution and use of 
 *  the Software in source and binary forms, with or without modification, 
 *  provided that the following conditions are met: 
 *  - Redistributions of source code must retain the above copyright notice, 
 *  this copyright license and the following disclaimer. 
 *  - Redistributions in binary form must reproduce the above copyright 
 *   notice, this copyright license and the following disclaimer in the 
 *  documentation and/or other materials provided with the distribution. 
 *  - Neither the name of LiPS  nor the names of its Members may be used 
 *  to endorse or promote products derived from the Software without 
 *  specific prior written permission. 
 *
 *  A patent license for any Necessary Claims owned by Members of LiPS Forum 
 *  to make, have made, use, import, offer to sell, lease and sell or otherwise 
 *  distribute any implementation compliant with the any specification adopted 
 *  by the LiPS Forumcan be obtained from the respective Members on reasonable 
 *  and non-discriminatory terms and conditions and under reciprocity, as 
 *  regulated in more detail in the Internal Policy of the LiPS Forum. 
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER, ITS MEMBERS AND CONTRIBUTORS 
 *  "AS IS", AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE 
 *  AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER, 
 *  ITS MEMBERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *  POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <uuid/uuid.h>
#include <glib.h>
#include <dirent.h>

#include <sqlite3.h>

#include <caleventbook.h>
#include "event-private.h"

#define UID_LEN      16
#define TIME_LENGTH    30

#define ATD_BASE "/var/spool/at"
#define TRIGGER ATD_BASE "/trigger"
/**
* Four tables are used in calendar. 
* 1.basic table: calendar
*    In this table, all events' information except category and exceptions are stored.Each event only 
*    has one row.
* 2.exception table: event_exception
*    In this table,the repeating events' exceptions are stored.
*    Each repeating event can have several rows to indicate several exceptions.
* 3.category table: event_category
*    In this table,the categories that events belong to are stored.One event can belong to several
*    categories.Therefore, one event can have several rows in this table.
* 4.alarm table: event_alarm
*    In this table, the alarm id, event id correspoding to the alarm, and the alarm time is stored.
*/
#define EVENT_TABLE_CREATION_SQL  "create table if not exists calendar" \
                                  " (tid0 BLOB NOT NULL," \
                                  " tid1 TEXT," \
                                  " tid2 TEXT," \
                                  " tid3 TEXT," \
                                  " tid4 INTEGER," \
                                  " tid5 INTEGER," \
                                  " tid6 INTEGER," \
                                  " tid7 INTEGER," \
                                  " tid8 INTEGER," \
                                  " tid9 INTEGER,"\
                                  " tid10 INTEGER," \
                                  " tid11 INTEGER," \
                                  " tid12 INTEGER," \
                                  " tid13 INTEGER," \
                                  " tid14 INTEGER," \
                                  " tid15 INTEGER," \
                                  " tid16 INTEGER," \
                                  " PRIMARY KEY(tid0)" \
                                  ");"
#define EVENT_EXCEPTION_CREATION_SQL        "create table if not exists event_exception"\
                                            " (uid BLOB NOT NULL,"\
                                            " exception INTEGER);"
#define EVENT_CATEGORY_CREATION_SQL          "create table if not exists event_category"\
                                            " (uid BLOB NOT NULL,"\
                                            " cid BLOB);"
#ifdef HAVE_AT
#define EVENT_ALARM_CREATION_SQL            "create table if not exists event_alarm" \
                                            " (aid INTEGER, " \
                                            " uid BLOB, " \
                                            " atime INTEGER, " \
                                            " sum  TEXT, "  \
                                            " PRIMARY KEY(aid));"
#else
#define EVENT_ALARM_CREATION_SQL            "create table if not exists event_alarm" \
                                            " (atime INTEGER, " \
                                            " uid BLOB, " \
                                            " sum  TEXT, "  \
                                            " PRIMARY KEY(atime));"
#endif


#define INSERT_BASIC_EVENT_SQL              "insert into calendar (tid0,tid1,"\
                                            "tid2,tid3,tid4,tid5,tid6,"\
                                            "tid7,tid8,tid9,tid10,tid11,"\
                                            "tid12,tid13,tid14,tid15,tid16) values"\
                                            " (?, '%q','%q','%q',"\
                                            "%d,%d,%d,"\
                                            "%d,%d,%d,"\
                                            "%d,%d,%d,"\
                                            "%d,%d,%d,%d);"
#define INSERT_EVENT_EXCEPTION_SQL          "insert into event_exception values" \
                                            " (?, %d);"
#ifdef HAVE_AT
#define INSERT_EVENT_ALARM_SQL              "insert into event_alarm values" \
                                            "  (%d, ?, %d, '%q');"
#else
#define INSERT_EVENT_ALARM_SQL              "insert into event_alarm values" \
                                            "  (%d, ?, '%q');"
#endif
#define DELETE_EVENT_SQL                    "delete from calendar where tid0 = ?;"
#define DELETE_EXCEPTION_SQL                "delete from event_exception where uid = ?;"
#define DELETE_ALARM_SQL                    "delete from event_alarm where uid = ?;"
#define CLEAR_EVENT_SQL                      "delete from calendar;"

static GString *sql_str = NULL;

struct _cal_eventbook_t
{
  guint32 id;
  gchar *db_file;
  sqlite3 *database;
};

typedef struct _QueryResult
{
  cal_event_iter_t *iterator;
  cal_event_t *event;
  cal_time_t from;
  cal_time_t to;
} QueryResult;


static gchar *
generate_uid_str (void)
{
  uuid_t uuid;
  gchar *str;

  str = g_new0 (gchar, UID_LEN + 1);
  uuid_generate (uuid);

  memcpy (str, uuid, 16);

  return str;
}

static cal_err_t
build_field_insert_string (cal_event_t *event)
{
  gchar *curr_str = NULL;

/**
 *  Check event validity
 */
  if (event->start > event->end)
  {
    g_print ("event->start should be lesser or equal to event->end");
    return CAL_ERROR_DATA_INVALID;
  }

  curr_str = sqlite3_mprintf (INSERT_BASIC_EVENT_SQL,
                              event->description,
                              event->location,
                              event->notes,
                              (gint) event->start,
                              (gint) event->end,
                              (gint) event->has_alarm,
                              event->alarm_delay,
                              event->alarm_type,
                              event->repeat_type,
                              event->repeat_frequency,
                              (gint) event->repeat_forever,
                              (gint) event->repeat_end,
                              event->repeat_weekdays,
                              event->repeat_monthdays,
                              (gint) event->repeat_on_nearest,
                              (gint) event->has_exception);
  g_string_assign (sql_str, curr_str);
  /*printf ("%s: sql_str=%s\n", __FUNCTION__, curr_str);*/
  sqlite3_free (curr_str);

  return ERROR_NONE;
}

static int
write_event_1 (sqlite3_stmt *pstmt, cal_event_t *event)
{
  gint i, tid;
  gchar *value = NULL;
  GString *colname = g_string_new (NULL);

/**
 * Put all the info about one event retrived from data storage into "tempev".
 */
  for (i = 0; i < sqlite3_column_count (pstmt); i++)
  {
    g_string_assign (colname, sqlite3_column_name (pstmt, i));
    g_string_erase (colname, 0, 3);
    tid = (gint) atoi (colname->str);
    /*g_print("tid = %d\n", tid);*/
    tid += CAL_TEMPLATE_ID_START;

    switch (tid)
    {
    case EVENT_UID:
      value = (gchar *) sqlite3_column_blob (pstmt, i);
      _event_set_uid (event, value);
      break;

    case EVENT_DESCRIPTION:
      value = (gchar *) sqlite3_column_text (pstmt, i);
      /*g_print("eventbook.c:%s: value=%s\n", __FUNCTION__, value);*/
      event_set_description (event, value);
      break;

    case EVENT_LOCATION:
      value = (gchar *) sqlite3_column_text (pstmt, i);
      /*g_print("eventbook.c:select_callback:\n value=%s\n", value);*/
      event_set_location (event, value);
      break;

    case EVENT_NOTES:
      value = (gchar *) sqlite3_column_text (pstmt, i);
      /*g_print("eventbook.c:select_callback:\n value=%s\n", value);*/
      event_set_notes (event, value);
      break;

    case EVENT_START_DATETIME:
      {
        int time;
        time = sqlite3_column_int (pstmt, i);
        event_set_start (event, (cal_time_t) time);
        break;
      }
    case EVENT_END_DATETIME:
      {
        int time;
        time = sqlite3_column_int (pstmt, i);
        event_set_end (event, (cal_time_t) time);
        break;
      }
    case EVENT_HAS_ALARM:
      {
        cal_boolean_t setting;
        setting = sqlite3_column_int (pstmt, i);
        event_set_has_alarm (event, setting);
        break;
      }
    case EVENT_ALARM_DELAY:
      event_set_alarm_delay (event, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_ALARM_TYPE:
      event_set_alarm_type (event, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_REPEAT_TYPE:
      event_set_repeat_type (event, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_REPEAT_FREQUENCY:
      event_set_repeat_frequency (event, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_REPEAT_FOREVER:
      event_set_repeat_forever (event, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_REPEAT_END_DATE:
      {
        int time;
        time = sqlite3_column_int (pstmt, i);
        event_set_repeat_end (event, (cal_time_t) time);
        break;
      }
    case EVENT_REPEAT_WEEKDAYS:
      event_set_repeat_weekdays (event, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_REPEAT_MONTHDAYS:
      event_set_repeat_monthdays (event, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_HAS_EXCEPTION:
      event_set_has_exception (event, sqlite3_column_int (pstmt, i));
      break;
    }                           /*switch (tid) */
  }                             /*for*/
  g_string_free (colname, TRUE);
  return ERROR_NONE;
}

static int
write_event (cal_eventbook_t *book, sqlite3_stmt *pstmt, void *user_data)
{
  QueryResult *result = (QueryResult *) user_data;
  gint i, tid;
  gchar *value = NULL;
  gchar *curr_str = NULL;
  gchar *uid_str;
  gint ret;
  cal_boolean_t match = FALSE;
  GString *colname = g_string_new (NULL);
  cal_event_t *tempev = event_new ();

  g_assert (tempev);
/**
 * Put all the info about one event retrived from data storage into "tempev".
 */
  for (i = 0; i < sqlite3_column_count (pstmt); i++)
  {
    g_string_assign (colname, sqlite3_column_name (pstmt, i));
    g_string_erase (colname, 0, 3);
    tid = atoi (colname->str);
    /*g_print("tid = %d\n", tid);*/
    tid += CAL_TEMPLATE_ID_START;

    switch (tid)
    {
    case EVENT_UID:
      value = (gchar *) sqlite3_column_blob (pstmt, i);
      /*g_print("eventbook.c:select_callback:\n value=%s\n", value);*/
      _event_set_uid (tempev, value);
      break;

    case EVENT_DESCRIPTION:
      value = (gchar *) sqlite3_column_text (pstmt, i);
      /*g_print("eventbook.c:select_callback:\n value=%s\n", value);*/
      event_set_description (tempev, value);
      break;

    case EVENT_LOCATION:
      value = (gchar *) sqlite3_column_text (pstmt, i);
      /*g_print("eventbook.c:select_callback:\n value=%s\n", value);*/
      event_set_location (tempev, value);
      break;

    case EVENT_NOTES:
      value = (gchar *) sqlite3_column_text (pstmt, i);
      /*g_print("eventbook.c:select_callback:\n value=%s\n", value);*/
      event_set_notes (tempev, value);
      break;

    case EVENT_START_DATETIME:
      {
        int time;
        time = sqlite3_column_int (pstmt, i);
        event_set_start (tempev, (cal_time_t) time);
        break;
      }
    case EVENT_END_DATETIME:
      {
        int time;
        time = sqlite3_column_int (pstmt, i);
        event_set_end (tempev, (cal_time_t) time);
        break;
      }
    case EVENT_HAS_ALARM:
      {
        cal_boolean_t setting;
        setting = sqlite3_column_int (pstmt, i);
        event_set_has_alarm (tempev, setting);
        break;
      }
    case EVENT_ALARM_DELAY:
      event_set_alarm_delay (tempev, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_ALARM_TYPE:
      event_set_alarm_type (tempev, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_REPEAT_TYPE:
      event_set_repeat_type (tempev, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_REPEAT_FREQUENCY:
      event_set_repeat_frequency (tempev, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_REPEAT_FOREVER:
      event_set_repeat_forever (tempev, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_REPEAT_END_DATE:
      {
        int time;
        time = sqlite3_column_int (pstmt, i);
        event_set_repeat_end (tempev, (cal_time_t) time);
        break;
      }
    case EVENT_REPEAT_WEEKDAYS:
      event_set_repeat_weekdays (tempev, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_REPEAT_MONTHDAYS:
      event_set_repeat_monthdays (tempev, sqlite3_column_int (pstmt, i));
      break;

    case EVENT_HAS_EXCEPTION:
      event_set_has_exception (tempev, sqlite3_column_int (pstmt, i));
      break;
    }                           /*switch (tid)*/
  }                             /*for (i = 0; i < result.nr_fields; i++)*/
  if (tempev->has_exception)
  {
    curr_str =
      sqlite3_mprintf ("select exception from event_exception where uid = ?;");
    ret = sqlite3_prepare (book->database, curr_str, -1, &pstmt, 0);
    uid_str = g_memdup (tempev->uid, 16);
    ret = sqlite3_bind_blob (pstmt, 1, uid_str, 16, free);
    while ((ret = sqlite3_step (pstmt)) == SQLITE_ROW)
    {
      cal_time_t exception;
      exception = sqlite3_column_int (pstmt, 0);
      event_add_exception (tempev, exception);
    }
    ret = sqlite3_finalize (pstmt);
    sqlite3_free (curr_str);
  }
/**
*  Check "tempev" to see whether it match the time limit "result->from" to "result->to" 
*/
  if (result->from == 0)
    match = TRUE;
  else
    match = _event_check_for_peroid (tempev, result->from, result->to);
/**
*  If match,according to "result->nr_fields" and "result->fields", copy the requested fields from tempev to 
*  result->event.
*/
  if (match)
  {
    result->event = g_memdup (tempev, sizeof (cal_event_t));
    _event_iterator_insert (result->iterator, result->event);
    g_assert (result->event);
  }

  g_string_free (colname, TRUE);
  if (!match)
    event_free (tempev);
  return ERROR_NONE;
}

cal_eventbook_t *
eventbook_connect (cal_uint32_t id)
{
  cal_eventbook_t *book = NULL;
  gchar *err_str;
  gint ret;
  gchar *res_path;

/*open database file and  create table if it's not exist*/
  book = g_new0 (cal_eventbook_t, 1);
  g_assert (book);

  if (!(res_path = getenv ("CAL_RES_PATH")))
  {
    g_print ("The env CAL_RES_PATH is not set!\n");
    return NULL;
  }
  book->db_file = g_build_filename (res_path, "calendar.db", NULL);
  g_assert (book->db_file);

  if (sqlite3_open (book->db_file, &book->database))
  {
    g_print ("eventbook: open database failed. %s\n",
             sqlite3_errmsg (book->database));
    sqlite3_close (book->database);
    g_free (book->db_file);
    g_free (book);

    return NULL;
  }
  ret =
    sqlite3_exec (book->database, EVENT_TABLE_CREATION_SQL, NULL, NULL,
                  &err_str);
  ret =
    sqlite3_exec (book->database, EVENT_EXCEPTION_CREATION_SQL, NULL, NULL,
                  &err_str);
  ret =
    sqlite3_exec (book->database, EVENT_CATEGORY_CREATION_SQL, NULL, NULL,
                  &err_str);
  ret =
    sqlite3_exec (book->database, EVENT_ALARM_CREATION_SQL, NULL, NULL,
                  &err_str);

#if 0
  if (ret != SQLITE_OK)
  {
    g_print ("eventbook: crreate table failed. [%s]\n", err_str);
    g_free (err_str);
    return CAL_ERROR_DB_TRANS;
  }
#endif
  return book;
}


void
eventbook_disconnect (cal_eventbook_t *book)
{
  sqlite3_close (book->database);
  g_free (book->db_file);
  g_free (book);

  return;
}

cal_err_t
_eventbook_add_exception (cal_eventbook_t *book, const cal_char_t *uid,
                          cal_time_t exception)
{
  sqlite3_stmt *pstmt = 0;
  gchar *curr_str = NULL;
  gchar *uid_str1, *uid_str2;
  cal_event_t *ev;
  gint ret;

  uid_str1 = g_memdup (uid, 16);
  uid_str2 = g_memdup (uid, 16);

  curr_str = sqlite3_mprintf (INSERT_EVENT_EXCEPTION_SQL, (gint) exception);
#ifdef DEBUG
  printf ("%s:%s\n", __FUNCTION__, curr_str);

  _print_uid (uid);
#endif
  ret = sqlite3_prepare (book->database, curr_str, -1, &pstmt, 0);
  g_print
    ("calendar: eventbook_add_event: add event exception prepare status:%d. \n",
     ret);
  ret = sqlite3_bind_blob (pstmt, 1, uid_str1, 16, free);
  g_print
    ("calendar: eventbook_add_event: event exception bind blob status:%d. \n",
     ret);
  ret = sqlite3_step (pstmt);
  g_print ("calendar: eventbook_add_event: event exception step status:%d. \n",
           ret);
  ret = sqlite3_finalize (pstmt);
  g_print
    ("calendar: eventbook_add_event: event exception finalize status:%d. \n",
     ret);

  sqlite3_free (curr_str);

  eventbook_get_event_by_uid (book, &ev, uid);
  /*Skip this occurrence's alarm */
  if (ev->has_alarm)
  {
    time_t oldatime, atime;
    struct tm oldtm, exceptm;

    localtime_r (&exception, &exceptm);
    oldatime = ev->start;
    localtime_r (&oldatime, &oldtm);
    exceptm.tm_sec = oldtm.tm_sec - ev->alarm_delay;
    exceptm.tm_min = oldtm.tm_min;
    exceptm.tm_hour = oldtm.tm_hour;
    atime = mktime (&exceptm);
#ifdef DEBUG
    printf ("%s : atime = %d\n", __FUNCTION__, atime);
#endif
    schedule_next (book, atime, uid);

  }
  return ERROR_NONE;
}

cal_err_t
eventbook_add_event (cal_eventbook_t *book, cal_event_t *event)
{
  gchar *uuid_str = NULL;
  sqlite3_stmt *pstmt = 0;
  gint ret;
  cal_err_t errno;

  uuid_str = generate_uid_str ();
  _event_set_uid (event, uuid_str);
  sql_str = g_string_new ("");
  errno = build_field_insert_string (event);

  if (errno != ERROR_NONE)
  {
    g_string_free (sql_str, TRUE);
    return errno;
  }
  else
  {
    ret = sqlite3_prepare (book->database, sql_str->str, -1, &pstmt, 0);
    /*g_print("calendar: eventbook_add_event: sqlite prepare status:%d. \n", ret);*/
    ret = sqlite3_bind_blob (pstmt, 1, uuid_str, 16, free);
    /*g_print("calendar: eventbook_add_event: sqlite bind blob status:%d. \n",ret);*/
    ret = sqlite3_step (pstmt);
    /*g_print("calendar: eventbook_add_event: sqlite step status:%d\n", ret);*/
    ret = sqlite3_finalize (pstmt);
    /*g_print("calendar: eventbook_add_event: sqlite finalize status:%d\n", ret);*/
  }
  g_string_free (sql_str, TRUE);
  /*
   * For new event we should call "_schedule_add_alarm" directly.But for repeat events,
   * to add exception to repeat event, we call "_eventbook_add_exception", in which ,the 
   * function "schedule_next" will call the function "_schedule_add_alarm"
   */
  if ((event->repeat_type != REPEAT_NONE) && (event->has_exception))
  {
    gint i;
    for (i = 0; i < event->exceptions->len; i++)
    {
      cal_time_t *exception;
      exception =
        g_ptr_array_index (event->exceptions, event->exceptions->len - i - 1);
#ifdef DEBUG
      printf ("%s:exception =%d\n", __FUNCTION__, *exception);
#endif
      _eventbook_add_exception (book, event->uid, *exception);
    }

  }
  else if (event->has_alarm)
  {
    cal_time_t alarm_time;
    struct tm *tm;

    alarm_time = event->start;

    if (event->alarm_delay != 0)
    {
      tm = localtime (&alarm_time);
      tm->tm_sec = tm->tm_sec - event->alarm_delay;
      alarm_time = mktime (tm);
    }
    _schedule_add_alarm (book, event->uid, alarm_time, event->description);
  }


  return ERROR_NONE;
}

cal_err_t
eventbook_remove_event (cal_eventbook_t *book, cal_char_t *uid)
{

  sqlite3_stmt *pstmt = 0;
  gchar *uuid_str1;
  gchar *uuid_str2;
  cal_event_t *ev;
  gint ret;

  g_assert (book != NULL);
  g_assert (uid != NULL);

  uuid_str1 = g_memdup (uid, 16);
  uuid_str2 = g_memdup (uid, 16);

#ifdef DEBUG
  printf ("enter %s \n", __FUNCTION__);
  _print_uid (uid);
#endif

  eventbook_get_event_by_uid (book, &ev, uid);

  ret = sqlite3_prepare (book->database, DELETE_EVENT_SQL, -1, &pstmt, 0);
  g_print ("calendar: eventbook_remove_event: sqlite prepare status:%d. \n",
           ret);
  ret = sqlite3_bind_blob (pstmt, 1, uuid_str1, 16, free);
  ret = sqlite3_step (pstmt);
  g_print ("eventbook_remove_event:sqlite3_step: ret=%d\n", ret);
  ret = sqlite3_finalize (pstmt);

  if ((ev->repeat_type != REPEAT_NONE) && (ev->has_exception))
  {
    ret = sqlite3_prepare (book->database, DELETE_EXCEPTION_SQL, -1, &pstmt, 0);
    g_print ("calendar: eventbook_remove_event: sqlite prepare status:%d. \n",
             ret);
    ret = sqlite3_bind_blob (pstmt, 1, uuid_str2, 16, free);
    ret = sqlite3_step (pstmt);
    g_print ("eventbook_remove_event:sqlite3_step: ret=%d\n", ret);
    ret = sqlite3_finalize (pstmt);
  }

/*
 * For the events which has alarm, we need to cancel the alarm. Be careful that ,if the 
 * alarm has been triggered, then the alarm record has been deleted from the event_alarm
 * table.
 */
  if (ev->has_alarm)
  {
    _schedule_cancel_alarm (book, uid);
  }


  return ERROR_NONE;
}


int scan_i = 0;
DIR *scan_dir = NULL;
char *dirs[] = { "/var/spool/at", NULL };

void
scan_from_top (void)
{
  if (scan_dir)
  {
    closedir (scan_dir);
    scan_dir = NULL;
    scan_i = 0;
  }
}

struct dirent *
scan (void)
{
  struct dirent *dirent;

  if (!scan_dir)
  {
    if (!dirs[scan_i])
    {
      scan_i = 0;
      return NULL;
    }
    else
      scan_dir = opendir (dirs[scan_i++]);
  }
  dirent = readdir (scan_dir);
  if (dirent)
    return dirent;
  closedir (scan_dir);
  scan_dir = NULL;

  return scan ();
}




cal_err_t
eventbook_remove_all_evnts (cal_eventbook_t *book)
{
  gint i;

  cal_event_iter_t *eviter = event_iterator_new ();

  eventbook_get_events_for_period (book, eviter, 0, 0);

  for (i = 0; i < event_iterator_size (eviter); i++)
  {
    cal_event_t *ev;

    ev = event_iterator_current (eviter);
    eventbook_remove_event (book, ev->uid);
    event_iterator_next (eviter);
  }

  return ERROR_NONE;

}

cal_err_t
eventbook_remove_events (cal_eventbook_t *book, cal_uint32_t nr_uids,
                         const cal_char_t *uids[])
{

  if (nr_uids != 0)
  {
    cal_int_t i;
    for (i = 0; i < nr_uids; i++)
    {
      eventbook_remove_event (book, (cal_char_t *) uids[i]);
    }
    return ERROR_NONE;
  }
  else
    return CAL_ERROR_DATA_INVALID;

}

cal_err_t
eventbook_get_event_by_uid (cal_eventbook_t *book,
                            cal_event_t **event, const cal_char_t *uid)
{
  sqlite3_stmt *pstmt = 0;
  gchar *curr_str;
  gchar *uid_str1;
  gchar *uid_str2;
  gint ret;


  *event = event_new ();

  if (*event == NULL)
  {
    return CAL_ERROR_MEM_ALLOC;
  }


  uid_str1 = g_memdup (uid, 16);
  uid_str2 = g_memdup (uid, 16);


  curr_str = sqlite3_mprintf ("select * from calendar where tid0 = ?;");
  ret = sqlite3_prepare (book->database, curr_str, -1, &pstmt, 0);
  ret = sqlite3_bind_blob (pstmt, 1, uid_str1, 16, free);
  while ((ret = sqlite3_step (pstmt)) == SQLITE_ROW)
  {
    write_event_1 (pstmt, *event);
  }

  ret = sqlite3_finalize (pstmt);
  sqlite3_free (curr_str);

  curr_str =
    sqlite3_mprintf ("select exception from event_exception where uid = ?;");
  ret = sqlite3_prepare (book->database, curr_str, -1, &pstmt, 0);
  ret = sqlite3_bind_blob (pstmt, 1, uid_str2, 16, free);
  while ((ret = sqlite3_step (pstmt)) == SQLITE_ROW)
  {
    cal_time_t exception;
    exception = sqlite3_column_int (pstmt, 0);
    event_add_exception (*event, exception);
  }
  ret = sqlite3_finalize (pstmt);
  sqlite3_free (curr_str);


  return ERROR_NONE;
}


/* "From " and "To " should be on the same day*/
cal_err_t
eventbook_get_events_for_period (cal_eventbook_t *book,
                                 cal_event_iter_t *iter,
                                 cal_time_t from, cal_time_t to)
{
  GString *condition_str = NULL;
  QueryResult result;
  gchar *err_str = NULL;
  gint ret;
  sqlite3_stmt *pstmt = 0;

  result.from = from;
  result.to = to;

  sql_str = g_string_new ("select * from calendar ");
  condition_str = g_string_new ("");
/*Build up condition_str*/
  if (from != 0)
  {
    if (to != 0)
    {

      condition_str->str =
        sqlite3_mprintf ("where ( tid9 = 0 AND tid4 < %d AND tid5 >= %d) OR "
                         "(tid9 !=0 AND tid4 < %d AND "
                         "(tid12 = 0 OR tid12 >= %d));", to, from, to, from);
    }
    else
    {
      condition_str->str = sqlite3_mprintf ("where (tid9 = 0 AND tid5 >= %d) OR"
                                            "(tid9!=0 AND (tid12 = 0 OR tid12 >= %d));",
                                            from, from);
    }
  }
  else
  {
    if (to != 0)
    {
      condition_str->str = sqlite3_mprintf ("where tid4 < %d;", to);
    }
    else
    {
      condition_str->str = sqlite3_mprintf (";");
    }
  }
  g_string_append (sql_str, condition_str->str);
  /*g_print("eventbook.c: eventbook_get_events_for_period:\n Query = \n[%s]\n", sql_str->str);*/

  result.iterator = iter;
  result.event = NULL;

  ret = sqlite3_prepare (book->database, sql_str->str, -1, &pstmt, 0);
  /*g_print("calendar: eventbook_get_event: sqlite prepare status:%d. \n", ret);*/

  while ((ret = sqlite3_step (pstmt)) == SQLITE_ROW)
  {
    /*g_print("calendar: eventbook_get_event: sqlite step status:%d. \n", ret);*/
    write_event (book, pstmt, &result);
  }
  ret = sqlite3_finalize (pstmt);

  g_string_free (sql_str, TRUE);
  g_string_free (condition_str, TRUE);

  if (ret != SQLITE_OK)
  {
    g_print ("calendar: get events failed %s\n", err_str);
    g_free (err_str);
    return CAL_ERROR_DB_TRANS;
  }

  return ERROR_NONE;
}

cal_err_t
eventbook_get_all_events (cal_eventbook_t *book, cal_event_iter_t *iter)
{
  gint ret;
  ret = eventbook_get_events_for_period (book, iter, 0, 0);
  return ret;
}

cal_err_t
eventbook_update_event (cal_eventbook_t *book, cal_event_t *event)
{
  gint ret;
  eventbook_remove_event (book, event->uid);

  ret = eventbook_add_event (book, event);
  return ret;
}


#ifdef HAVE_AT
static gchar *
alarm_filename ()
{
#ifdef DEBUG
  printf ("/tmp/atjob.txt\n");
#endif
  return g_strdup_printf ("/tmp/atjob.txt");
}
#else
static gchar *
alarm_filename (time_t start)
{
  gchar *temp;
  uid_t uid = getuid ();

  temp = g_strdup_printf ("%s/%ld.%ld", ATD_BASE, start, uid);
#ifdef DEBUG
  /*printf ("%s:%s\n", __FUNCTION__, temp);*/
#endif
  return temp;
}
#endif

#ifndef HAVE_AT
static gboolean
trigger_atd (void)
{
  gchar pid_file[25];
  gint fd;
  gboolean rc = TRUE;

#ifdef DEBUG
  printf ("trigger_atd\n");
#endif
  /*sprintf (pid_file, "/var/run/atd.pid");
     if (access (pid_file, F_OK) != 0)
     rc =  FALSE; */
  if (rc)
  {
#ifdef DEBUG
    printf ("open trigger file and write\n");
#endif
    fd = open (TRIGGER, O_WRONLY);
    if (fd < 0)
    {
      perror (TRIGGER);
      return FALSE;
    }

    if (write (fd, "1\n", 2) < 0)
      rc = FALSE;
    if (close (fd) < 0)
      rc = FALSE;
  }
  return TRUE;
}
#endif
#ifdef HAVE_AT
static const char *boilerplate_1 =
  "#!/bin/sh\nexport DISPLAY=:1\nexport LD_LIBRARY_PATH=/usr/local/lib\n";
#else
static const char *boilerplate_1 = "#!/bin/sh\nexport DISPLAY=:0\n";
#endif
/*static const char * condensorplate_1 = "\nrm -f $0\n";*/

cal_boolean_t
_schedule_add_alarm (cal_eventbook_t * book, const cal_char_t * uid, cal_time_t time,
                     const cal_char_t * description)
{
  gint fd;
#ifdef HAVE_AT
  gchar *filename = alarm_filename ();
  struct tm *tm;
  gchar call_at[256];
  FILE *at_return;
#else
  gchar *filename = alarm_filename (time);
#endif
  FILE *fp;
  gchar *action;
  gint alarm_id = 0;
  sqlite3_stmt *pstmt = 0;
  gchar *curr_str = NULL;
  gchar *uid_str;
  gint ret;

  uid_str = g_memdup (uid, 16);
#ifdef DEBUG
  printf ("eventbook.c:_schedule_add_alarm. time=%d, description=%s\n",
          (int) time, description);
#endif
  fd = open (filename, O_WRONLY | O_CREAT, 0700);
  if (fd < 0)
  {
    g_free (filename);
    g_free (uid_str);
    return FALSE;
  }
  if ((fp = fdopen (fd, "w")) == NULL)
  {
#ifdef DEBUG
    printf ("can't reopen atjob file\n");
#endif
    return FALSE;
  }

  fwrite (boilerplate_1, sizeof (char), strlen (boilerplate_1), fp);
  action =
    g_strdup_printf
    ("announce '%s'\nesdplay $CAL_RES_PATH/alarm.wav\nschedule -s %ld \n",
     description, time);
  fwrite (action, sizeof (char), strlen (action), fp);
  fclose (fp);
  g_free (filename);
#ifdef HAVE_AT
  tm = localtime (&time);
  sprintf (call_at,
           "/usr/bin/at -q g -f /tmp/atjob.txt %02d:%02d %02d.%02d.%02d 2>&1",
           tm->tm_hour, tm->tm_min, tm->tm_mday, tm->tm_mon + 1,
           tm->tm_year - 100);
  g_print ("%s\n", call_at);
  if ((at_return = popen (call_at, "r")) != NULL)
  {
    gchar junk[4];
    do
    {
      if (!fscanf (at_return, "%s ", junk))
        break;
    } while (strcmp (junk, "job") != 0);
    if (fscanf (at_return, "%i", &alarm_id))
    {
#ifdef DEBUG
      printf ("alarm_id =%d\n", alarm_id);
#endif
    }
    pclose (at_return);

    if (alarm_id != 0)
    {

      curr_str =
        sqlite3_mprintf (INSERT_EVENT_ALARM_SQL, alarm_id, (gint) time,
                         description);
      ret = sqlite3_prepare (book->database, curr_str, -1, &pstmt, 0);
      g_print
        ("calendar: eventbook_add_event: add event alarm prepare status:%d. \n",
         ret);
      ret = sqlite3_bind_blob (pstmt, 1, uid_str, 16, free);
      g_print
        ("calendar: eventbook_add_event: event alarm bind blob status:%d. \n",
         ret);
      ret = sqlite3_step (pstmt);
      g_print ("calendar: eventbook_add_event: event alarm step status:%d. \n",
               ret);
      ret = sqlite3_finalize (pstmt);
      g_print
        ("calendar: eventbook_add_event: event alarm finalize status:%d. \n",
         ret);
      sqlite3_free (curr_str);

    }
  }
  else
  {
#ifdef DEBUG
    printf ("ERROR:Date string was invalid or couldn't run 'at'.\n");
#endif
    return FALSE;
  }
  return TRUE;

#else


  curr_str = sqlite3_mprintf (INSERT_EVENT_ALARM_SQL, (gint) time, description);
  ret = sqlite3_prepare (book->database, curr_str, -1, &pstmt, 0);
  g_print
    ("calendar: eventbook_add_event: add event alarm prepare status:%d. \n",
     ret);
  ret = sqlite3_bind_blob (pstmt, 1, uid_str, 16, free);
  g_print ("calendar: eventbook_add_event: event alarm bind blob status:%d. \n",
           ret);
  ret = sqlite3_step (pstmt);
  g_print ("calendar: eventbook_add_event: event alarm step status:%d. \n",
           ret);
  ret = sqlite3_finalize (pstmt);
  g_print ("calendar: eventbook_add_event: event alarm finalize status:%d. \n",
           ret);
  sqlite3_free (curr_str);

  trigger_atd ();
  return TRUE;
#endif

}


cal_boolean_t
_schedule_cancel_alarm (cal_eventbook_t *book, const cal_char_t *uid)
{
  gchar *uid_str1;
  sqlite3_stmt *pstmt = 0;
  gchar *curr_str = NULL;
  gint ret;
#ifdef HAVE_AT
  gint ret_val;
  gint aid;
  gchar erase_at[256];
  gchar **uid_str2;
#else
  gint atime;
#endif
#ifdef HAVE_AT
  uid_str1 = g_memdup (uid, 16);
  uid_str2 = g_memdup (uid, 16);

  curr_str = sqlite3_mprintf ("select aid from event_alarm where uid = ?;");
  ret = sqlite3_prepare (book->database, curr_str, -1, &pstmt, 0);
  ret = sqlite3_bind_blob (pstmt, 1, uid_str1, 16, free);
  ret = sqlite3_step (pstmt);
  aid = sqlite3_column_int (pstmt, 0);

  ret = sqlite3_finalize (pstmt);
  sqlite3_free (curr_str);


  sprintf (erase_at, "/usr/bin/atrm %d", aid);
  ret_val = system (erase_at);
  g_print ("%s:ret_val =%d\n", __FUNCTION__, ret_val);

  ret = sqlite3_prepare (book->database, DELETE_ALARM_SQL, -1, &pstmt, 0);
  /*g_print("calendar: eventbook_remove_event: sqlite prepare status:%d. \n", ret);*/
  ret = sqlite3_bind_blob (pstmt, 1, uid_str2, 16, free);
  ret = sqlite3_step (pstmt);
  /*g_print("eventbook_remove_event:sqlite3_step: ret=%d\n", ret);*/
  ret = sqlite3_finalize (pstmt);
#else
  uid_str1 = g_memdup (uid, 16);

  curr_str = sqlite3_mprintf ("select atime from event_alarm where uid = ?;");


#ifdef DEBUG
  printf ("%s:curr_str=%s\n", __FUNCTION__, curr_str);
#endif
  ret = sqlite3_prepare (book->database, curr_str, -1, &pstmt, 0);
  ret = sqlite3_bind_blob (pstmt, 1, uid_str1, 16, free);
  ret = sqlite3_step (pstmt);
  atime = sqlite3_column_int (pstmt, 0);
#ifdef DEBUG
  printf ("atime = %d\n", atime);
#endif
  ret = sqlite3_finalize (pstmt);
  sqlite3_free (curr_str);

  if (atime != 0)
  {
    curr_str =
      sqlite3_mprintf ("delete from event_alarm where atime = %d;", atime);
#ifdef DEBUG
    printf ("%s:curr_str=%s\n", __FUNCTION__, curr_str);
#endif
    sqlite3_exec (book->database, curr_str, NULL, NULL, NULL);
    sqlite3_free (curr_str);

    gchar *filename = alarm_filename (atime);
#ifdef DEBUG
    printf ("%s:filename=%s\n", __FUNCTION__, filename);
#endif
    if (unlink (filename))
    {
      g_free (filename);
      return FALSE;
    }
    g_free (filename);

    return trigger_atd ();
  }
#endif
  return FALSE;
}

void
schedule_next (cal_eventbook_t *book, cal_time_t skip, const cal_char_t *evid)
{
  sqlite3_stmt *pstmt = 0;
  gchar *curr_str = NULL;
  cal_event_t *event = NULL;
  gint alarm_id = 0;
  gint ret;

  if ((!evid) && (skip != 0))
  {
    gchar *tempevid = NULL;
  /**
  * When the time reach for the alarm.Find the item in event_alarm where time =  skip.
  */
#ifdef HAVE_AT
    curr_str =
      sqlite3_mprintf ("select aid,uid from event_alarm where atime = %d;",
                       skip);
#else
    curr_str =
      sqlite3_mprintf ("select uid from event_alarm where atime = %d;", skip);
#endif
#ifdef DEBUG
    printf ("%s:%s\n", __FUNCTION__, curr_str);
#endif
    ret = sqlite3_prepare (book->database, curr_str, -1, &pstmt, 0);
    while ((ret = sqlite3_step (pstmt)) == SQLITE_ROW)
    {
#ifdef HAVE_AT
      alarm_id = sqlite3_column_int (pstmt, 0);
#endif
      tempevid = g_memdup ((gchar *) sqlite3_column_blob (pstmt, 0), 16);

      /*tempevid = (gchar *) sqlite3_column_text(pstmt, 0);*/
#ifdef DEBUG
      g_print ("--evid = ");
      _print_uid (tempevid);
#endif
    }
    ret = sqlite3_finalize (pstmt);

    sqlite3_free (curr_str);


    if (tempevid)
    {

      eventbook_get_event_by_uid (book, &event, tempevid);

   /**
    * Find the event in calendar according to event uid.If the event 
    * is a repeating event, calculate the next alarm and 
    * update the table event_alarm and add alarm.
    */
#ifdef HAVE_AT
      curr_str =
        sqlite3_mprintf ("delete from event_alarm where aid = %d ;", alarm_id);

#else
      curr_str =
        sqlite3_mprintf ("delete from event_alarm where atime = %d ;", skip);

#ifdef DEBUG
      printf ("%s:%s\n", __FUNCTION__, curr_str);
#endif
      ret = sqlite3_prepare (book->database, curr_str, -1, &pstmt, 0);
      ret = sqlite3_step (pstmt);
      ret = sqlite3_finalize (pstmt);
      sqlite3_free (curr_str);

      gchar *filename = alarm_filename (skip);

#ifdef DEBUG
      printf ("%s:filename=%s\n", __FUNCTION__, filename);
#endif
      if (unlink (filename))
      {
        g_free (filename);
        return;
      }
      g_free (filename);

#endif

      if (event_get_repeat_type (event) != REPEAT_NONE)
      {
        cal_time_t next_alarm_time;
        cal_boolean_t ok = FALSE;
#ifdef DEBUG
        printf ("%s:event repeat \n", __FUNCTION__);
#endif
        next_alarm_time = event_get_next_occurence (event, skip, &ok);
#ifdef DEBUG
        printf ("next_alarm_time=%d\n", (int) next_alarm_time);
        printf ("ok = %d\n", ok);
#endif
        if (ok)
        {
          struct tm *tm;
          if (event_get_has_alarm (event))
          {
            tm = localtime (&next_alarm_time);
            tm->tm_sec -= event_get_alarm_delay (event);
            next_alarm_time = mktime (tm);
          }
          _schedule_add_alarm (book, evid, next_alarm_time,
                               event_get_description (event));
        }
      }

      event_free (event);
      g_free (tempevid);
    }
  }

  else if (evid)
  {
    eventbook_get_event_by_uid (book, &event, evid);
    if (event_get_repeat_type (event) != REPEAT_NONE)
    {
      cal_time_t next_alarm_time;
      cal_boolean_t ok = FALSE;
#ifdef DEBUG
      printf ("%s:event repeat \n", __FUNCTION__);
#endif
      _schedule_cancel_alarm (book, evid);
      next_alarm_time = event_get_next_occurence (event, skip, &ok);
#ifdef DEBUG
      printf ("next_alarm_time=%d\n", (int) next_alarm_time);
      printf ("ok = %d\n", ok);
#endif
      if (ok)
      {
        _schedule_add_alarm (book, evid, next_alarm_time,
                             event_get_description (event));
      }
    }

    event_free (event);
  }
}

/*
vi:ts=2:nowrap:ai:expandtab
*/

